unit MobiUnit;

interface

uses
  SysUtils, Windows, Classes, System.Types, PalmDatabaseRO;

type
  TPointerReadOnlyStream = class(TCustomMemoryStream)
  protected
    function Write(const Buffer; Count: Longint): Longint; reintroduce;
  public
    constructor Create(P: Pointer; Size: NativeInt);
    procedure SetPointer(Ptr: Pointer; const Size: NativeInt);
  end;

function SwapLong(Value: Cardinal): Cardinal;

function GetIdNamePosFromStream(const Stream: TStream; MaxMBytesProceed: Integer; idname: array of Byte; FromPos: Int64 = 0): Int64;

function GetMobiCover(const FileName: string; FirstImageIfNoCover: Boolean): TBytes;
function GetMobiCoverIdxFromEXTH(const Stream: TStream): Integer;

implementation

//https://wiki.mobileread.com/wiki/MOBI

type
  TMobiExthHeader = packed record
    Magic: array [0..3] of AnsiChar;
    Length: Integer;
    RecordCount: Integer;
  end;

  TMobiExthRecord = packed record
    RecType: Integer;
    RecLength: Integer;
  end;

const
  MobiHeaderMagic: AnsiString = 'MOBI';
  MobiExthHeaderMagic: AnsiString = 'EXTH';


function SwapLong(Value: Cardinal): Cardinal;
//https://stackoverflow.com/questions/5133938/procedure-that-swaps-the-bytes-low-high-of-a-word-variable
asm
  {$IFDEF CPUX64}
  mov rax, rcx
  {$ENDIF}
  bswap eax
end;

function GetIdNamePosFromStream(const Stream: TStream; MaxMBytesProceed: Integer; idname: array of Byte; FromPos: Int64 = 0): Int64;
var
  BTMP: TBytes;
  I: Int64;
  MS: TCustomMemoryStream;
begin
  Result := -1;
  if Length(idname) = 0 then
    Exit;

  if FromPos <=0 then
    I := -1
  else
    I := FromPos - 1;

  SetLength(BTMP, Length(idname));

  if Stream is TCustomMemoryStream then
  begin
    MS := TCustomMemoryStream(Stream);

    while I < Stream.Size - Length(idname) do
    begin
      Inc(I);
      if PByte(NativeUInt(MS.Memory) + I)^ = idname[0] then
        if CompareMem(Pointer(NativeUInt(MS.Memory) + I), @idname[0], Length(idname)) then
        begin
          Result := I;
          Break;
        end;

      if MaxMBytesProceed > 0 then
        if I >= (MaxMBytesProceed * 1024 * 1024) then
          Break;
    end;
  end else
  while I < Stream.Size - Length(idname) do
  begin
    Inc(I);
    Stream.Position := I;
    Stream.ReadBuffer(BTMP[0], Length(BTMP));
    if CompareMem(@BTMP[0], @idname[0], Length(idname)) then
    begin
      Result := I;
      Break;
    end;

    if MaxMBytesProceed > 0 then
      if Stream.Position >= (MaxMBytesProceed * 1024 * 1024) then
        Break;
  end;
end;

function GetMobiCover(const FileName: string; FirstImageIfNoCover: Boolean): TBytes;
var
  Mobi: TPalmPDB;
  PS: TPointerReadOnlyStream;
  B: TBytes;
  Magic: array [0..3] of AnsiChar;
  CoverIndex, ImageIndex: Integer;
  ExthPos: Int64;
begin
  Result := nil;
  CoverIndex := -1;

  Mobi := TPalmPDB.CreateReadOnly(FileName);
  try
    B := Mobi.GetRecordB(0);
    if Length(B) = 0 then
      Exit;

    PS := TPointerReadOnlyStream.Create(Pointer(B), Length(B));
    try
      PS.Position := 16;
      PS.ReadBuffer(Magic, Length(Magic));
      if Magic <> MobiHeaderMagic then
        Exit;
      PS.Position := 108;
      PS.Read(ImageIndex, SizeOf(ImageIndex));
      ImageIndex := SwapLong(ImageIndex);
      if ImageIndex <= 0 then // -1? But zero offset for images seems impossible
        Exit;

      ExthPos := GetIdNamePosFromStream(PS, 1, [Ord('E'), Ord('X'), Ord('T'), Ord('H')], PS.Position);
      if ExthPos > 0 then
      begin
        PS.Position := ExthPos;
        CoverIndex := GetMobiCoverIdxFromEXTH(PS);
       //MessageBox(GetActiveWindow, PChar(IntToStr(CoverIndex) + #13 + IntToStr(CoverIndex)), nil, 0);
      end;

      if CoverIndex = -1 then
      begin
        if FirstImageIfNoCover then
          CoverIndex := 0 // just first image
        else
          Exit;
      end;

      ImageIndex := ImageIndex + CoverIndex;

      if (ImageIndex < 0) or (Cardinal(ImageIndex) >= Cardinal(Mobi.RecordListCount)) then
        Exit;

      Result := Mobi.GetRecordB(ImageIndex);
    finally
      PS.Free;
    end;
  finally
    Mobi.Free;
  end;
end;

function GetMobiCoverIdxFromEXTH(const Stream: TStream): Integer;
var
  Header: TMobiExthHeader;
  Records: Integer;
  Rec: TMobiExthRecord;
  I: Integer;
begin
  Result := -1;
  Stream.Read(Header, SizeOf(Header));
  if Header.Magic <> MobiExthHeaderMagic then
    Exit;
  Records := SwapLong(Header.RecordCount);
  if Records <= 0 then
    Exit;

  try
    for I := 0 to Records - 1 do
    begin
      Stream.ReadBuffer(Rec, SizeOf(Rec));
      Rec.RecType := SwapLong(Rec.RecType);
      Rec.RecLength := SwapLong(Rec.RecLength) - SizeOf(Rec);
      if Rec.RecType = 201 then
      begin
        Stream.ReadBuffer(Result, SizeOf(Result));
        Result := SwapLong(Result);
      end else
        Stream.Seek(Rec.RecLength, soFromCurrent);
    end;
  except
    Result := -1;
  end;
end;


{ TPointerStream }

constructor TPointerReadOnlyStream.Create(P: Pointer; Size: NativeInt);
begin
  SetPointer(P, Size);
  Position := 0;
end;

procedure TPointerReadOnlyStream.SetPointer(Ptr: Pointer; const Size: NativeInt);
begin
  inherited SetPointer(Ptr, Size);
end;

function TPointerReadOnlyStream.Write(const Buffer; Count: Integer): Longint;
begin
  raise EStreamError.Create('Not implemented');
end;

end.
