{
  Changed sources from
  https://totalcmd.net/plugring/PdbView.html
  https://totalcmd.net/plugring/PdbView_2.html
}
unit PalmDatabaseRO;

interface

uses SysUtils, Classes;

const
  MAX_DB_NAME_LENGTH = 31;

type
  /// ///////////////////////////////////////////////////////////
  // Record type for the database header (common to PDB and PRC)
  TPDBHeaderNative = packed record
    Name: array [0 .. MAX_DB_NAME_LENGTH] of AnsiChar;
    Flags, Version: Word;
    Created, Modified, BackedUp: Longint;
    Modification, AppInfoArea, SortInfoArea: Longint;
    DBType, CreatorID: array [0 .. 3] of AnsiChar;
    IDSeed, NextRecordList: Longint;
    NumRecords: Word;
  end;

  /// ///////////////////////////////////////////////////////////
  // Record type for the record list (PDB)
  TPDBRecordHeaderNative = packed record
    DataOffset: Longint;
    Flags: AnsiChar;
    UniqueID: array [0 .. 2] of AnsiChar;
  end;

  // try this eventually
  // TDBFlag = (dbReadOnly, dbDirty, dbBackup,
  // dbOverwrite, dbForceReset, dbNoBeam);
  // TDBFlags = set of TDBFlag;

  /// ///////////////////////////////////////////////////////////
  // TPDBHeader
  //
  // The database header (for PDB and PRC)
  TPDBHeader = record
    Name: AnsiString;
    ReadOnly, Dirty, Backup, Overwrite, ForceReset, NoBeam: Boolean;
    // DBFlags: TDBFlags;
    Version: Integer;
    Created, Modified, BackedUp: TDateTime;
    Modification, AppInfoArea, SortInfoArea: Integer;
    DBType, CreatorID: AnsiString;
    IDSeed, NextRecordList: Integer;
    NumRecords: Integer;
  end;

  TRecFlag = (rfSecret, rfBusy, rfDirty, rfDelete);
  TRecFlags = set of TRecFlag;

  /// ///////////////////////////////////////////////////////////
  // TPDBRecordHeader
  //
  // Record header class (A PDB maintains a TList of these)
  // Physically, this is right after the database header in the file
  TPDBRecordHeader = class
  public
    Offset: Integer;
    RecFlags: TRecFlags;
    UniqueID: array [0 .. 2] of AnsiChar;
    Size: Integer;
    RecordData: PAnsiChar;
    constructor Create(Offset, Size: Integer);
    destructor Destroy; override;
  end;

  /// ///////////////////////////////////////////////////////////
  // TPalmPDB
  //
  // PDB database
  TPalmPDB = class
  strict private
    Stream: TStream;
    procedure GetHeader(const Stream: TStream);
    procedure InternalFillHeaderData(var DBHeaderNative: TPDBHeaderNative);
  private
    DBHeader: TPDBHeader;
    DBFileSize: Integer;
    PDBFile: string;
    PDBRecordList: TList;
    procedure GetRecordList(const Stream: TStream);
    procedure ClearRecordList;
    function GetRecordListCount: Integer;
  public
    constructor CreateReadOnly(const FileName: string);
    destructor Destroy; override;
    function GetRecordB(RecNo: Integer): TBytes;
    class function SwapLongInt(L: Longint): Longint;
    class function PalmToTDateTime(Seconds: Cardinal): TDateTime;
    property GetDBHeader: TPDBHeader read DBHeader;
    property RecordListCount: Integer read GetRecordListCount;
  end;


  EInvalidPalmDatabase = class(Exception);

  /// ///////////////////////////////////////////////////////////
  /// ///////////////////////////////////////////////////////////
  // Implementation
  /// ///////////////////////////////////////////////////////////
  /// ///////////////////////////////////////////////////////////

implementation

/// ///////////////////////////////////////////////////////////
// TPDBRecordHeader definitions
//
constructor TPDBRecordHeader.Create(Offset, Size: Integer);
begin
  if (Offset < 0) or (Size < 0) then
    raise EInvalidPalmDatabase.Create
      ('Invalid Palm database: record headers corrupt');
  Self.Offset := Offset;
  Self.Size := Size;
end;

// Testing
destructor TPDBRecordHeader.Destroy;
begin
  inherited Destroy;
end;

/// ///////////////////////////////////////////////////////////
// TPalmPDB definitions
//
constructor TPalmPDB.CreateReadOnly(const FileName: string);
begin
  // Initialize the header class so we have access to it later
  PDBFile := FileName;
  PDBRecordList := TList.Create;

  Stream := TFileStream.Create(PDBFile, fmOpenRead or fmShareDenyWrite);
  Stream.Position := 0;
  GetHeader(Stream);
  GetRecordList(Stream);
end;

destructor TPalmPDB.Destroy;
begin
  try
    // Free the data pointed to by the record list
    ClearRecordList;

    // Free the list
    PDBRecordList.Free;
  finally
    Stream.Free;
  end;

  // Free everything else
  inherited Destroy;
end;

procedure TPalmPDB.GetHeader(const Stream: TStream);
var
  DBHeaderNative: TPDBHeaderNative;
begin
  try
    DBFileSize := Stream.Size;
    Stream.ReadBuffer(DBHeaderNative, Sizeof(TPDBHeaderNative));
  except
    raise EInvalidPalmDatabase.Create('Error reading file');
  end;

  InternalFillHeaderData(DBHeaderNative);
end;

procedure TPalmPDB.InternalFillHeaderData(var DBHeaderNative: TPDBHeaderNative);
begin
  try
    // Assign the native types to Delphi types
    DBHeader.Name := DBHeaderNative.Name;
    // Remember, the bytes are swapped on Intel
    DBHeaderNative.Flags := Swap(DBHeaderNative.Flags);
    DBHeader.ReadOnly := Boolean(DBHeaderNative.Flags and $0002);
    DBHeader.Dirty := Boolean(DBHeaderNative.Flags and $0004);
    DBHeader.Backup := Boolean(DBHeaderNative.Flags and $0008);
    DBHeader.Overwrite := Boolean(DBHeaderNative.Flags and $0010);
    DBHeader.ForceReset := Boolean(DBHeaderNative.Flags and $0020);
    DBHeader.NoBeam := Boolean(DBHeaderNative.Flags and $0040);
    DBHeader.Version := Swap(DBHeaderNative.Version);
    DBHeader.Created := PalmToTDateTime(SwapLongInt(DBHeaderNative.Created));
    DBHeader.Modified := PalmToTDateTime(SwapLongInt(DBHeaderNative.Modified));
    DBHeader.BackedUp := PalmToTDateTime(SwapLongInt(DBHeaderNative.BackedUp));
    DBHeader.Modification := SwapLongInt(DBHeaderNative.Modification);
    DBHeader.AppInfoArea := SwapLongInt(DBHeaderNative.AppInfoArea);
    DBHeader.SortInfoArea := SwapLongInt(DBHeaderNative.SortInfoArea);
    DBHeader.DBType := DBHeaderNative.DBType;
    DBHeader.CreatorID := DBHeaderNative.CreatorID;
    DBHeader.IDSeed := SwapLongInt(DBHeaderNative.SortInfoArea);
    DBHeader.NextRecordList := SwapLongInt(DBHeaderNative.SortInfoArea);
    DBHeader.NumRecords := Swap(DBHeaderNative.NumRecords);
  except
    raise EInvalidPalmDatabase.Create('Not a Palm database, or header corrupt');
  end;
end;

// Convert the Motorola format Lontint to the Intel format
class function TPalmPDB.SwapLongInt(L: Longint): Longint;
begin
  Result := Swap(L shr 16) or (Longint(Swap(L and $FFFF)) shl 16);
end;

// Delphi determines the date/time based on days since 12/30/1899
// Palm determines the date/time based on seconds since 1/1/1904
class function TPalmPDB.PalmToTDateTime(Seconds: Cardinal): TDateTime;
var
  Days, Remainder: Integer;
  // This function doesn't work right
  // It was translating 1999 dates to 1933, not sure why, this ErrorOffset
  // kludge fixes it pretty well... ErrorOffset gets us to 1/1/1970
  // (Unix time).  I don't question it, I just use it...
const
  ErrorOffset = 2082844800; // seconds
begin
  Inc(Seconds, ErrorOffset);
  // 1462 is the number of days between 12/30/1899 and 1/1/1904
  Inc(Seconds, (1462 * 86400));
  Days := Seconds div 86400;
  Remainder := Seconds mod 86400;
  Result := Days + (Remainder / 86400);
end;

procedure TPalmPDB.GetRecordList(const Stream: TStream);
var
  I: Integer;
  PDBRecordHeaderNative: TPDBRecordHeaderNative;
begin
  // Skip past the header
  Stream.Position := Sizeof(TPDBHeaderNative);

  for I := 0 to DBHeader.NumRecords - 1 do
  begin
    // Read off a record header
    Stream.ReadBuffer(PDBRecordHeaderNative, Sizeof(TPDBRecordHeaderNative));

    // Add it to the list
    PDBRecordList.Add(TPDBRecordHeader.Create
      (SwapLongInt(PDBRecordHeaderNative.DataOffset), 0));
    TPDBRecordHeader(PDBRecordList.Items[I]).RecordData := nil;

    // Determine the size of the previous record (now that we know this one's offset)
    if PDBRecordList.Count > 1 then
      TPDBRecordHeader(PDBRecordList.Items[PDBRecordList.Count - 2]).Size :=
        TPDBRecordHeader(PDBRecordList.Last).Offset -
        TPDBRecordHeader(PDBRecordList.Items[PDBRecordList.Count - 2]).Offset;
  end;

  // Set the size of the last one
  TPDBRecordHeader(PDBRecordList.Last).Size := Stream.Size -
    TPDBRecordHeader(PDBRecordList.Last).Offset;
end;

function TPalmPDB.GetRecordListCount: Integer;
begin
  Result := PDBRecordList.Count;
end;

procedure TPalmPDB.ClearRecordList;
var
  I: Integer;
begin
  // Free all the header items
  PDBRecordList.Pack;
  for I := 0 to PDBRecordList.Count - 1 do
  begin
    // Free any data still associated with the record
    if TPDBRecordHeader(PDBRecordList.Items[I]).RecordData <> nil then
      StrDispose(TPDBRecordHeader(PDBRecordList.Items[I]).RecordData);
    // Free the header item
    TPDBRecordHeader(PDBRecordList.Items[I]).Free;
  end;
  PDBRecordList.Clear;
end;

function TPalmPDB.GetRecordB(RecNo: Integer): TBytes;
begin
  if (RecNo < 0) or (Cardinal(RecNo) >= Cardinal(PDBRecordList.Count)) then
    raise EInvalidPalmDatabase.Create('Invalid record index: ' + IntToStr(RecNo));

  SetLength(Result, TPDBRecordHeader(PDBRecordList.Items[RecNo]).Size);
  if Length(Result) = 0 then
    Exit;

  // Skip to the requested record's offset
  Stream.Position := TPDBRecordHeader(PDBRecordList.Items[RecNo]).Offset;

  // Read the record
  Stream.ReadBuffer(Result[0], TPDBRecordHeader(PDBRecordList.Items[RecNo]).Size); // Or just Stream.Read ?
end;

end.
